"""
Build optical system and analyze.

"""


import numpy as np 
from . import util 
import matplotlib.pyplot as plt

system_verbosity = True


class System(object):
    """
    Entire optical system.
    
    Attributes
    ----------
    matrix : ndarray
        cascaded beam transformation matrix
    f : ndarray/float
        frequency
    wout : ndarray/float
        output beam waist
    qout : ndarray/complex
        output beam parameter
    rout : ndarray/float
        output beam radius
    
    """

    def __init__(self, horn_tx, component_list, horn_rx=None, **kwargs):
        """
        System constructor.
        
        Parameters
        ----------
        horn_tx : class
            transmitting horn
        component_list : tuple/list
            list of optical components
        horn_rx : class
            receiving horn
        \*\*kwargs : 
            keyword arguments (comment, verbose)
        
        """

        if horn_rx is None:
            horn_rx = horn_tx.copy(comment='copy', verbose=False)

        assert horn_rx.f == horn_tx.f, "Horn frequencies must match."
        self.f = horn_rx.f

        self.comment = kwargs.get('comment', '')
        self._verbose = kwargs.get('verbose', system_verbosity)

        self._horn_tx = horn_tx
        self._horn_rx = horn_rx
        self._comp_list = component_list

        self._system = component_list[0]
        for comp in component_list[1:]:
            self._system *= comp 
        self.matrix = self._system.matrix

        self.qout = transform_beam(self.matrix, self._horn_tx.q)
        self.wout = _waist_from_q(self._horn_tx.q, self.f.w)
        self.rout = _radius_from_q(self._horn_tx.q)

        if self._verbose:
            print((self.__str__()))

    def __str__(self):

        return "System: {0}\n{1}".format(self.comment, self.matrix)

    def coupling(self):
        """
        Get coupling between the horns.
        
        Returns
        -------
        ndarray 
            coupling between the antennas
            
        """

        return _coupling(self.qout, self._horn_rx, self.f.w)

    def best_coupling_frequency(self):
        """
        Get best coupling frequency.
        
        Returns
        -------
        float
            frequency where best coupling is found
        
        """

        idx_best = self.coupling().argmax()

        return self.f.f[idx_best]

    def best_coupling(self):
        """
        Get best coupling.
        
        Returns
        -------
        float
            best coupling 
            
        """

        return self.coupling().max()

    def print_best_coupling(self, units='GHz'):
        """
        Print best coupling and frequency where it is found.
        
        Parameters
        ----------
        units : str, optional
            units for frequency

        """

        mult = util.set_f_units(units)
        best = self.best_coupling() * 100.
        f_best = self.best_coupling_frequency() / mult

        s = "Best coupling: {0:.1f} % at {1:.1f} {2}"
        print((s.format(best, f_best, units)))

    def plot_coupling(self, ax=None):
        """
        Plot coupling (in percentage) versus frequency.
        
        """

        f = self.f.f / self.f.mult

        if ax is None:
            fig, ax = plt.subplots()
        ax.plot(f, self.coupling() * 100.)
        ax.set(xlim=[f.min(), f.max()])
        ax.set(xlabel='Frequency ({0})'.format(self.f.units))
        ax.set(ylabel='Coupling (\%)')
        plt.autoscale(enable=True, axis='y', tight=True)
        plt.grid(True)
        plt.minorticks_on()
        if ax is None:
            plt.show()

    def plot_coupling_mag(self, ax=None):
        """
        Plot coupling versus frequency. 

        """

        f = self.f.f / self.f.mult

        if ax is None:
            fig, ax = plt.subplots()
        ax.plot(f, self.coupling())
        ax.set(xlim=[f.min(), f.max()])
        ax.set(xlabel='Frequency ({0})'.format(self.f.units))
        ax.set(ylabel='Coupling (1)')
        if ax is None:
            plt.show()

    def plot_coupling_db(self, ax=None):
        """
        Plot coupling (in dB) versus frequency.

        """

        f = self.f.f / self.f.mult

        if ax is None:
            fig, ax = plt.subplots()
        ax.plot(f, 10*np.log10(self.coupling()))
        ax.set(xlim=[f.min(), f.max()])
        ax.set(xlabel='Frequency ({0})'.format(self.f.units))
        ax.set(ylabel='Coupling (dB)')
        if ax is None:
            plt.show()

    def plot_edge_taper_db(self, ax=None):

        f = self.f.f / self.f.mult

        if ax is None:
            fig, ax = plt.subplots()
        sys = self._comp_list[0]
        for comp in self._comp_list[1:]:
            sys *= comp
            if comp.type == 'obj':
                q = transform_beam(sys.matrix, self._horn_tx.q)
                w = _waist_from_q(q, self.f.w)
                et = _edge_taper(w, comp.radius, output='dB')
                ax.plot(f, et, label=comp.comment)
        ax.legend()
        ax.set(xlim=[f.min(), f.max()])
        ax.set(xlabel='Frequency ({0})'.format(self.f.units))
        ax.set(ylabel='Edge Taper (dB)')
        plt.tight_layout()
        if ax is None:
            plt.show()

    def plot_waists(self, ax=None):

        f = self.f.f / self.f.mult

        if ax is None:
            fig, ax = plt.subplots()
        sys = self._comp_list[0]
        for comp in self._comp_list[1:]:
            sys *= comp
            if comp.type == 'obj':
                q = transform_beam(sys.matrix, self._horn_tx.q)
                w = _waist_from_q(q, self.f.w) * 1e3
                ax.plot(f, w, label=comp.comment)
        ax.legend()
        ax.set(xlim=[f.min(), f.max()])
        ax.set(xlabel='Frequency ({0})'.format(self.f.units))
        ax.set(ylabel='Beam Waist (mm)')
        if ax is None:
            plt.show()

    def plot_aperture_30db(self, ax=None):

        f = self.f.f / self.f.mult

        if ax is None:
            fig, ax = plt.subplots()
        sys = self._comp_list[0]
        for comp in self._comp_list[1:]:
            sys *= comp
            if comp.type == 'obj':
                q = transform_beam(sys.matrix, self._horn_tx.q)
                w = _waist_from_q(q, self.f.w) * 1e3
                ax.plot(f, w*np.sqrt(30/8.686), label=comp.comment)
        ax.legend()
        ax.set(xlim=[f.min(), f.max()])
        ax.set(xlabel='Frequency ({0})'.format(self.f.units))
        ax.set(ylabel='Required Aperture Radius (mm)')
        plt.tight_layout()
        if ax is None:
            plt.show()

    def plot_system(self, freq=None, ax=None, figname=None):

        # Frequency to plot waists for
        if freq is None:
            idx = self.f.idx_center
        else:
            idx = self.f.idx(freq)

        qin = self._horn_tx.q
        wlen = self.f.w

        if ax is None:
            fig, ax = plt.subplots()

        # Plot waist as a function of distance
        d_last_comp = 0.
        distance = []
        waist = []
        sys = np.matrix([[1., 0], [0., 1.]])
        for comp in self._comp_list:
            if comp.type == 'prop':
                d_tmp = np.arange(comp.distance, step=1e-3)
                for d in d_tmp:
                    m_tmp = _freespace_matrix(d)
                    s_tmp = np.dot(m_tmp, sys)
                    q = transform_beam(s_tmp, qin)
                    w = _waist_from_q(q, wlen)
                    waist.append(w[idx])
                    distance.append(d + d_last_comp)
            d_last_comp += comp.distance
            sys = np.dot(comp.matrix, sys)
        distance = np.array(distance)
        waist = np.array(waist)
        ax.plot(distance*1e3, waist*1e3, label='Beam waist')
        ax.plot(distance*1e3, waist*np.sqrt(30/8.686)*1e3, 'r--',
                label='Aperture radius required\n'+ r'for $<$30 dB edge taper')

        # Add component radii and positions
        distance = self._comp_list[0].distance
        sys = self._comp_list[0]
        for comp in self._comp_list[1:]:
            sys *= comp
            distance += comp.distance
            if comp.type == 'obj':
                q = transform_beam(sys.matrix, self._horn_tx.q)
                w = _waist_from_q(q, self.f.w)

                ax.axvline(distance * 1e3, ls=':', lw=0.5, c='k')

                ax.plot(distance * 1e3, comp.radius * 1e3, 'kx', lw=1, markersize=3)
                label = "A.R.".format(comp.comment)
                ax.text(distance * 1e3, comp.radius * 1e3 + 4, label,
                        rotation=90, ha='center', va='bottom',
                        multialignment='left',
                        bbox=dict(boxstyle='round', facecolor='wheat',
                                  alpha=1),
                        fontsize=8)

                label = "{0}".format(comp.comment)
                ax.text(distance * 1e3, w[idx] * 1e3 + 10, label,
                        rotation=90, ha='center', va='bottom',
                        multialignment='left',
                        bbox=dict(boxstyle='round', facecolor='wheat',
                                  alpha=1),
                        fontsize=8)

        ax.set(ylim=[0, 100], ylabel='Size (mm)')
        plt.autoscale(enable=True, axis='x', tight=True)
        ax.set(xlabel='Distance from Horn Aperture (mm)')
        ax.minorticks_on()

        leg = ax.legend(loc=8, fontsize=8)
        leg.get_frame().set_alpha(1.)

        if figname is not None:
            plt.savefig(figname, bbox_inches='tight')
        else:
            plt.tight_layout()
            plt.show()


def transform_beam(sys_matrix, q_in):
    """
    Transform a beam using the beam transformation matrix.
    
    Parameters
    ----------
    sys_matrix : ndarray
        beam transformation matrix
    q_in : complex/ndarray
        input beam parameter (q)

    Returns
    -------
    complex/ndarray
        output beam parameter (q)

    """

    q_out = (sys_matrix[0, 0] * q_in + sys_matrix[0, 1]) / \
            (sys_matrix[1, 0] * q_in + sys_matrix[1, 1])

    return q_out


# Helper functions -----------------------------------------------------------

def _waist_from_q(q, wavelength):

    return np.sqrt(wavelength / (np.pi * np.imag(-1 / q)))


def _radius_from_q(q):

    return 1 / (np.real(1 / q))


def _freespace_matrix(distance):

    return np.matrix([[1., distance], [0., 1.]])


def _coupling(qin, horn_rx, wavelength):
    """Determine the coupling to a horn.

    Given the systems output beam parameter q, this function will determine
    the coupling to a given horn.

    Args:
        qin (complex): Input beam parameter
        horn_rx: Receiving horn
        wavelength (ndarray/float): Wavelength

    Returns:
        ndarray/float: Coupling (some value 0-1)

    """

    whorn = horn_rx.w
    zoff = horn_rx.z

    if isinstance(wavelength, np.ndarray):
        qout = np.zeros(np.alen(wavelength), dtype=complex)
        for i in range(np.alen(wavelength)):
            qout[i] = transform_beam(_freespace_matrix(zoff[i]), qin[i])
    else:
        qout = transform_beam(_freespace_matrix(zoff), qin)

    w = np.sqrt(wavelength / (np.pi * np.imag(-1 / qout)))

    r = 1 / (np.real(1 / qout))

    coupling = 4 / ((((w / whorn) + (whorn / w))**2 +
                     ((np.pi * whorn * w / wavelength)**2) *
                     (((1 / r) - (1 / 1e50))**2)))

    return coupling


def _edge_taper(beam_waist, aperture_radius, output='dB'):
    """Determine the edge taper due to some aperture.

    Args:
        beam_waist (ndarray): beam waist
        aperture_radius (float): aperture radius
        output (string): Return edge taper in 'dB' or 'linear' units

    Returns:
        ndarray: edge taper

    """

    taper = np.exp(-2 * aperture_radius ** 2 / beam_waist ** 2)

    if output == 'linear':
        return taper

    elif output == 'dB':
        return -10 * np.log10(taper)



"""
Class for frequency sweep.

"""


import numpy as np 
import scipy.constants as sc
from . import util

frequency_verbosity = True


class Frequency(object):
    """
    Class for frequency sweep.
    
    Attributes
    ----------
    f : float/ndarray
        frequency array (in Hz)
    w : float/ndarray
        wavelength array (in m)
    idx_center : int
        index of center value
    
    """

    def __init__(self, start=None, stop=None, npts=301, units='GHz', **kwargs):
        """
        Frequency constructor.
        
        Parameters
        ----------
        start : float, optional
            start frequency
        stop : float, optional
            stop frequency
        npts : int, optional
            number of points
        units : str, optional
            frequency units (e.g., 'GHz', 'MHz')
        \*\*kwargs : 
            keyword arguments
            center/span can be used to specify frequency
            single can be used to use a single frequency
            verbose and comment 
        
        """

        center = kwargs.pop('center', None)
        span = kwargs.pop('span', None)
        single = kwargs.pop('single', None)
        self._comment = kwargs.pop('comment', None)
        verbose = kwargs.pop('verbose', frequency_verbosity)

        self.units = units
        self.mult = util.set_f_units(units)

        if single is not None:
            start = single
            stop = single
            npts = 1
        elif center is not None and span is not None:
            start = center - span/2
            stop = center + span/2
        elif start is None and stop is None:
            start = 200
            stop = 300
        
        self.f = np.linspace(start, stop, npts) * self.mult
        self.w = sc.c / self.f

        self.idx_center = npts // 2

        if verbose:
            print((self.__str__()))

    def __str__(self):

        start = self.f[0] / self.mult
        stop = self.f[-1] / self.mult
        npts = np.alen(self.f)
        if self._comment is None:
            s = "Frequency sweep:\n\tf = {0:.1f} to {1:.1f} {2}, {3} pts\n"
            s = s.format(start, stop, self.units, npts)
        else:
            s = "Frequency sweep: {0}\n\tf = {1:.1f} to {2:.1f} {3}, {4} pts\n"
            s = s.format(self._comment, start, stop, self.units, npts)

        return s

    def __repr__(self):

        return self.__str__()

    def __len__(self):
        
        return np.alen(self.f)

    def __eq__(self, other):

        return np.array_equal(self.f, other.f)

    def idx(self, freq, units='GHz'):
        """
        Get index of value closest to specified frequency.
        
        Parameters
        ----------
        freq : float
            target frequency
        units : str 
            units for the frequency

        Returns
        -------
        int
            index of value closest to given frequency

        """

        mult = util.set_f_units(units)

        return np.abs(freq * mult - self.f).argmin()



"""
General utilities.

"""

import scipy.constants as sc
from . import component
from . import frequency
from . import system


def set_d_units(units):
    """
    Read distance units.

    Parameters
    ----------
    units : str 
        distance units (e.g., 'mm', 'um', 'cm')

    Returns
    -------
    float
        multiplier

    """

    unit_dict = {
        'km': sc.kilo,
        'm': 1,
        'dm': sc.deci,
        'cm': sc.centi,
        'mm': sc.milli,
        'um': sc.micro,
    }

    try:
        mult = unit_dict[units.lower()]
    except:
        raise ValueError("Units not recognized.")

    return mult


def set_f_units(units):
    """
    Read frequency units.

    Parameters
    ----------
    units : str 
        frequency units (e.g., 'THz', 'GHz', 'MHz')

    Returns
    -------
    float
        multiplier

    """

    unit_dict = {
        'thz': sc.tera,
        'ghz': sc.giga,
        'mhz': sc.mega,
        'khz': sc.kilo,
        'hz': 1,
    }

    try:
        mult = unit_dict[units.lower()]
    except:
        raise ValueError("Units not recognized.")

    return mult


def set_verbosity(verbosity):

    component.component_verbosity = verbosity
    frequency.frequency_verbosity = verbosity
    system.system_verbosity = verbosity



""" Analyze quasi-optical systems using Gaussian beam analysis.

"""



from . import frequency
from . import component
from . import system
from . import util

from .frequency import *
from .component import *
from .system import System 
from .util import set_verbosity

__author__ = "John Garrett"
__version__ = "0.1"



#!/usr/bin/env python

from distutils.core import setup

setup(name='GaussOpt',
      version='1.1.3',
      description='Gaussian beam analysis',
      keywords='gaussian optics millimeter terahertz thz',
      packages=['gaussopt',],
      author='John Garrett',
      author_email='garrettj403@gmail.com',
      license='MIT',
      url='https://github.com/garrettj403/GaussOpt/',
      long_description=open('README').read(),
     )



""" Classes for optical components.

"""


import numpy as np
import scipy.constants as sc
import matplotlib.pyplot as plt
from copy import deepcopy as copy
from .system import transform_beam
from . import util

component_verbosity = True


# Optical component base-class ------------------------------------------------

class Component(object):
    """
    Base-class for a generic component in a Gaussian optical system.
    
    To get initialization information, see :func:`__init__`.
    
    Attributes
    ----------
    matrix : ndarray
        beam transformation matrix, 2x2
    
    """

    def __init__(self, matrix=None, **kwargs):
        """
        Component constructor.
        
        Parameters
        ----------
        matrix : ndarray
            beam transformation matrix, 2x2
        \*\*kwargs : 
            key word arguments, such as 'comment', 'units', 'radius' and 
            'verbose'
            
        """

        # Set beam transformation matrix
        if matrix is not None:
            self.matrix = matrix
        else:
            self.matrix = np.matrix([[1., 0.], [0., 1.]])

        # Unpack keyword arguments
        self.comment = kwargs.get('comment', '')
        self._units = kwargs.get('units', 'mm')
        self._mult = util.set_d_units(self._units)
        self.radius = kwargs.get('radius', None)
        self._verbose = kwargs.get('verbose', component_verbosity)

        if self.radius is not None:
            self.radius *= self._mult

        # Set default values
        self.distance = 0.
        self.type = 'comp'

    def __str__(self):

        return 'New Component: {0}\n{1}\n'.format(self.comment, self.matrix)

    def __repr__(self):

        return self.__str__()

    def __mul__(self, next_comp):

        new_matrix = np.dot(next_comp.matrix, self.matrix)
        new_comp = Component(new_matrix, verbose=False,
                             comment='Composite System')

        return new_comp


# Propagation classes ---------------------------------------------------------

class Freespace(Component):
    """
    Freespace propagation.
    
    To get initialization information, see :func:`__init__`.
    
    Attributes
    ----------
    matrix : ndarray
        beam transformation matrix, 2x2
        
    """

    def __init__(self, distance, **kwargs):
        """
        Freespace constructor.
        
        Parameters
        ----------
        distance : float
            freespace propagation distance
        \*\*kwargs : 
            key word arguments, such as 'comment', 'units', 'radius' and 
            'verbose'
            
        """

        Component.__init__(self, **kwargs)

        self.type = 'prop'
        self.distance = distance * self._mult

        self.matrix = np.matrix([[1., self.distance], [0., 1.]])

        if self._verbose:
            print((self.__str__()))

    def __str__(self):

        string = "Freespace: {0}\n\td = {1:.1f} {2}\n"
        distance_red = self.distance / self._mult
        string = string.format(self.comment, distance_red, self._units)
        
        return string


class Dielectric(Component):
    """
    Propagation through a dielectric slab.
    
    To get initialization information, see :func:`__init__`.
    
    Attributes
    ----------
    matrix : ndarray
        beam transformation matrix, 2x2
        
    """

    def __init__(self, thickness, n=1., **kwargs):
        """
        Dielectric constructor.
        
        Parameters
        ----------
        thickness : float
            thickness of dielectric slab
        n : float, optional
            index of refraction
        \*\*kwargs : 
            key word arguments, such as 'comment', 'units', 'radius' and 
            'verbose'
        
        """

        Component.__init__(self, **kwargs)

        self.type = 'prop'
        self.distance = thickness * self._mult
        self._n = n

        self.matrix = np.matrix([[1., self.distance * n], [0., 1.]])

        if self._verbose:
            print((self.__str__()))

    def __str__(self):

        string = "Dielectric: {0}\n\td = {1:.1f} {2}, n = {3:.1f}\n"
        d_red = self.distance / self._mult
        string = string.format(self.comment, d_red, self._units, self._n)
        
        return string


# Transformation classes ------------------------------------------------------

class Mirror(Component):
    """
    Reflection off of a parabolic mirror.
    
    To get initialization information, see :func:`__init__`.
    
    Attributes
    ----------
    matrix : ndarray
        beam transformation matrix, 2x2
        
    """

    def __init__(self, focal_length, **kwargs):
        """
        Mirror constructor.
        
        Parameters
        ----------
        focal_length : float
            mirror focal length
        \*\*kwargs : 
            key word arguments, such as 'comment', 'units', 'radius' and 
            'verbose'
        
        """

        Component.__init__(self, **kwargs)

        self.type = 'obj'
        self._focal_length = focal_length * self._mult

        self.matrix = np.matrix([[1., 0.], [-1. / self._focal_length, 1.]])

        if self._verbose:
            print((self.__str__()))

    def __str__(self):

        string = "Mirror: {0}\n\tf = {1:.1f} {2}\n"
        f_red = self._focal_length / self._mult
        string = string.format(self.comment, f_red, self._units)
        
        return string


class Window(Component):
    """
    A window.
    
    To get initialization information, see :func:`__init__`.
    
    Attributes
    ----------
    matrix : ndarray
        beam transformation matrix, 2x2
        
    """

    def __init__(self, **kwargs):
        """
        Window constructor.
        
        Parameters
        ----------
        \*\*kwargs : 
            key word arguments, such as 'comment', 'units', 'radius' and 
            'verbose'
        
        """

        Component.__init__(self, **kwargs)

        self.type = 'obj'
    
    def __str__(self):

        if self.radius is None:
            string = "Window: {0}\n"
            string = string.format(self.comment)
        else:
            arad = self.radius * 1e3
            string = "Window: {0}\n\taperture radius = {1:.1f} mm\n"
            string = string.format(self.comment, arad)

        return string


# Horn ------------------------------------------------------------------------

class Horn(object):
    """
    Waveguide horn antenna.
    
    To get initialization information, see :func:`__init__`.
    
    Attributes
    ----------
    f : ndarray/float
        frequency
    w : ndarray/float
        beam waist at aperture
    z : ndarray/float
        z-offset
    q : ndarray/complex
        beam parameter at aperture
        
    """

    def __init__(self, freq, slen, arad, hf=0.59, **kwargs):
        """
        Horn constructor.
        
        Parameters
        ----------
        freq : class 'gaussopt.frequency.Frequency'
            frequency class
        slen : float
            slant length
        arad : float
            aperture radius
        hf : float, optional
            horn factor
        \*\*kwargs : 
            key word arguments, such as 'comment', 'units', and 'verbose'
        """

        self.type = 'horn'

        self.comment = kwargs.get('comment', '')
        self._units = kwargs.get('units', 'mm')
        self.radius = kwargs.get('radius', None)
        self._verbose = kwargs.get('verbose', component_verbosity)

        self._mult = util.set_d_units(self._units)

        slen *= self._mult
        arad *= self._mult

        self._slen = slen
        self._arad = arad
        self._hf = hf
        self.f = freq

        self.q, self.w, self.z = _horn(slen, arad, hf, freq.w)

        if self._verbose:
            print((self.__str__()))

    def __str__(self):

        slen = self._slen / self._mult
        arad = self._arad / self._mult
        units = self._units

        p1 = "\tslen = {0:5.2f} {1}\n".format(slen, units)
        p2 = "\tarad = {0:5.2f} {1}\n".format(arad, units)
        p3 = "\thf   = {0:5.2f}\n".format(self._hf)

        return 'Horn: {0}\n{1}{2}{3}'.format(self.comment, p1, p2, p3)

    def __repr__(self):

        return self.__str__()

    def z_offset(self, units='mm'):
        """
        Get distance between horn aperture and beam waist (a.k.a. z-offset).
        
        Parameters
        ----------
        units : str, optional
            units to use for returned value

        Returns
        -------
        float
            z offset

        """

        mult = util.set_d_units(units)

        return self.z / mult

    def waist(self, units='mm'):
        """
        Beam waist at aperture.
        
        Parameters
        ----------
        units : str, optional
            units to use for returned values

        Returns
        -------
        float
            waist at aperture

        """

        mult = util.set_d_units(units)

        return self.w / mult

    def copy(self, **kwargs):
        """
        Copy horn to new instance.
        
        Parameters
        ----------
        \*\*kwargs : 
            keyword arguments to pass to new instance

        Returns
        -------
        class
            new instance of the Horn class

        """

        tmp = copy(self)
        tmp.comment = kwargs.get('comment', '')
        tmp._verbose = kwargs.get('verbose', component_verbosity)
        if tmp._verbose:
            print((tmp.__str__()))

        return tmp

    def plot_properties(self):
        """
        Plots beam waist and z-offset over frequency range.

        """

        f = self.f.f / self.f._mult

        fig, ax = plt.subplots()
        ax.plot(f, self.w / self._mult, label='Waist at Aperture')
        ax.plot(f, self.z / self._mult, label='Z offset')
        ax.set(xlim=[f.min(), f.max()])
        ax.set(xlabel='Frequency ({0})'.format(self.f._units))
        ax.set(ylabel='Size ({0})'.format(self._units))
        plt.legend()
        plt.show()


def _horn(slen, arad, hf, wlen):
    """
    Calculate the beam parameters for a given horn.
    
    Parameters
    ----------
    slen : float
        slant length
    arad : float
        aperture radius
    hf : float
        horn factor
    wlen : float/ndarray
        wavelength array

    Returns
    -------
    tuple
        beam parameter, beam waist and z-offset

    """

    # wlen can either be an array or a float
    if type(wlen) != np.ndarray:

        # Eqn 7.38a and 7.38b (note there is an error in the textbook)
        w = hf * arad / np.sqrt(1 + (np.pi * (hf * arad) ** 2 /
                                (wlen * slen)) ** 2)
        zoff = slen / (1 + (wlen * slen / (np.pi * (hf * arad) ** 2)) ** 2)
        q_waist = _q_from_waist(w, wlen)

        waist_to_ap = Freespace(zoff, units='m', verbose=False).matrix
        qap = transform_beam(waist_to_ap, q_waist)

        return qap, w, zoff

    # A crappy hack, but this isn't a computer intensive package
    elif type(wlen) == np.ndarray:

        npts = np.alen(wlen)
        qap_arr = np.zeros_like(wlen, dtype=complex)
        w_arr = np.zeros_like(wlen)
        zoff_arr = np.zeros_like(wlen)

        for i in range(npts):

            qap_arr[i], w_arr[i], zoff_arr[i] = _horn(slen, arad, hf, wlen[i])

        return qap_arr, w_arr, zoff_arr

      
# Helper functions -----------------------------------------------------------

def _q_from_waist(waist, wlen):
    """
    Get the beam parameter (q) from the waist.
    
    Notes
    -----
    This is only valid for the case R=inf (i.e., at the beam waist).
    
    Parameters
    ----------
    waist : float
        beam waist
    wlen : float/ndarray
        wavelength array

    Returns
    -------
    ndarray
        beam parameter (q)

    """

    return 1j * sc.pi * waist ** 2 / wlen
